#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""SMOS L2 wind Reader.

Data can be found here after register:
https://www.smosstorm.org/Data2/SMOS-NRT-wind-Products-access
Format documentation at the same site after register:
SMOS_WIND_DS_PDD_20191107_signed.pdf
"""

import datetime as dt
import logging

import numpy as np
from pyresample.geometry import AreaDefinition

from satpy.readers.core.netcdf import NetCDF4FileHandler, netCDF4

logger = logging.getLogger(__name__)


class SMOSL2WINDFileHandler(NetCDF4FileHandler):
    """File handler for SMOS L2 wind netCDF files."""

    @property
    def start_time(self):
        """Get start time."""
        return dt.datetime.strptime(self["/attr/time_coverage_start"], "%Y-%m-%dT%H:%M:%S Z")

    @property
    def end_time(self):
        """Get end time."""
        return dt.datetime.strptime(self["/attr/time_coverage_end"], "%Y-%m-%dT%H:%M:%S Z")

    @property
    def platform_shortname(self):
        """Get platform shortname."""
        return self.filename_info["platform_shortname"]

    @property
    def platform_name(self):
        """Get platform."""
        return self["/attr/platform"]

    def get_metadata(self, data, ds_info):
        """Get metadata."""
        metadata = {}
        metadata.update(data.attrs)
        metadata.update(ds_info)
        metadata.update({
            "platform_shortname": self.platform_shortname,
            "platform_name": self.platform_name,
            "sensor": self["/attr/instrument"],
            "start_time": self.start_time,
            "end_time": self.end_time,
            "level": self["/attr/processing_level"],
        })

        return metadata

    def available_datasets(self, configured_datasets=None):
        """Automatically determine datasets provided by this file."""
        handled_variables = set()

        # Iterate over dataset contents
        for var_name, val in self.file_content.items():
            # Only evaluate variables
            if not isinstance(val, netCDF4.Variable):
                continue
            if (var_name in handled_variables):
                logger.debug("Already handled, skipping: %s", var_name)
                continue
            handled_variables.add(var_name)
            new_info = {
                "name": var_name,
                "file_type": self.filetype_info["file_type"],
            }
            yield True, new_info

    def _mask_dataset(self, data):
        """Mask out fill values."""
        try:
            fill = data.attrs["_FillValue"]
            data.attrs["_FillValue"] = np.nan
            return data.where(data != fill)
        except KeyError:
            return data

    def _adjust_lon_coord(self, data):
        """Adjust lon coordinate to -180 .. 180 ( not 0 .. 360)."""
        data = data.assign_coords(lon=(((data.lon + 180) % 360) - 180))
        return data.where(data < 180., data - 360.)

    def _rename_coords(self, data):
        """Rename coords."""
        rename_dict = {}
        if "lon" in data.dims:
            data = self._adjust_lon_coord(data)
            rename_dict["lon"] = "x"
        if "lat" in data.dims:
            rename_dict["lat"] = "y"
        # Rename the coordinates to x and y
        return data.rename(rename_dict)

    def _remove_time_coordinate(self, data):
        """Remove time coordinate."""
        # Remove dimension where size is 1, eg. time
        data = data.squeeze()
        # Remove if exists time as coordinate
        if "time" in data.coords:
            data = data.drop_vars("time")
        return data

    def _roll_dataset_lon_coord(self, data):
        """Roll dataset along the lon coordinate."""
        if "lon" in data.dims:
            data = data.roll(lon=720, roll_coords=True)
        return data

    def get_dataset(self, ds_id, ds_info):
        """Get dataset."""
        data = self[ds_id["name"]]
        data.attrs = self.get_metadata(data, ds_info)
        data = self._remove_time_coordinate(data)
        data = self._roll_dataset_lon_coord(data)
        data = self._rename_coords(data)
        data = self._mask_dataset(data)
        if len(data.dims) >= 2 and all([dim in data.dims for dim in ["x", "y"]]):
            # Remove the first and last row as these values extends beyond +-90 latitude
            # if the dataset contains the y dimmension.
            # As this is data over open sea these has no values.
            data = data.where((data.y > -90.0) & (data.y < 90.0), drop=True)
        elif len(data.dims) == 1 and "y" in data.dims:
            data = data.where((data.y > 0) & (data.y < len(data.y) - 1), drop=True)
        return data

    def _create_area_extent(self, width, height):
        """Create area extent."""
        # Creating a meshgrid, not needed actually, but makes it easy to find extremes
        _lon = self._adjust_lon_coord(self["lon"])
        _lon = self._roll_dataset_lon_coord(_lon)
        latlon = np.meshgrid(_lon, self["lat"][1:self["lat/shape"][0] - 1])
        lower_left_x = latlon[0][height - 1][0] - 0.125
        lower_left_y = latlon[1][height - 1][0] + 0.125
        upper_right_y = latlon[1][1][width - 1] - 0.125
        upper_right_x = latlon[0][1][width - 1] + 0.125
        return (lower_left_x, lower_left_y, upper_right_x, upper_right_y)

    def get_area_def(self, dsid):
        """Define AreaDefintion."""
        width = self["lon/shape"][0]
        height = self["lat/shape"][0] - 2
        area_extent = self._create_area_extent(width, height)
        description = "SMOS L2 Wind Equirectangular Projection"
        area_id = "smos_eqc"
        proj_id = "equirectangular"
        proj_str = self["/attr/geospatial_bounds_vertical_crs"]
        area_def = AreaDefinition(area_id, description, proj_id, proj_str, width, height, area_extent, )
        return area_def



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Writer for TIFF images compatible with the NinJo visualization tool (NinjoTIFFs).

NinjoTIFFs can be color images or monochromatic. For monochromatic images, the
physical units and scale and offsets to retrieve the physical values are
provided. Metadata is also recorded in the file.

In order to write ninjotiff files, some metadata needs to be provided to the
writer. Here is an example on how to write a color image::

    chn = "airmass"
    ninjoRegion = load_area("areas.def", "nrEURO3km")

    filenames = glob("data/*__")
    global_scene = Scene(reader="hrit_msg", filenames=filenames)
    global_scene.load([chn])
    local_scene = global_scene.resample(ninjoRegion)
    local_scene.save_dataset(chn, filename="airmass.tif", writer='ninjotiff',
                          sat_id=6300014,
                          chan_id=6500015,
                          data_cat='GPRN',
                          data_source='EUMCAST',
                          nbits=8)

Here is an example on how to write a color image::

    chn = "IR_108"
    ninjoRegion = load_area("areas.def", "nrEURO3km")

    filenames = glob("data/*__")
    global_scene = Scene(reader="hrit_msg", filenames=filenames)
    global_scene.load([chn])
    local_scene = global_scene.resample(ninjoRegion)
    local_scene.save_dataset(chn, filename="msg.tif", writer='ninjotiff',
                          sat_id=6300014,
                          chan_id=900015,
                          data_cat='GORN',
                          data_source='EUMCAST',
                          physic_unit='K',
                          nbits=8)

The metadata to provide to the writer can also be stored in a configuration file
(see pyninjotiff), so that the previous example can be rewritten as::

    chn = "IR_108"
    ninjoRegion = load_area("areas.def", "nrEURO3km")

    filenames = glob("data/*__")
    global_scene = Scene(reader="hrit_msg", filenames=filenames)
    global_scene.load([chn])
    local_scene = global_scene.resample(ninjoRegion)
    local_scene.save_dataset(chn, filename="msg.tif", writer='ninjotiff',
                          # ninjo product name to look for in .cfg file
                          ninjo_product_name="IR_108",
                          # custom configuration file for ninjo tiff products
                          # if not specified PPP_CONFIG_DIR is used as config file directory
                          ninjo_product_file="/config_dir/ninjotiff_products.cfg")


.. _ninjotiff: http://www.ssec.wisc.edu/~davidh/polar2grid/misc/NinJo_Satellite_Import_Formats.html

"""

import logging

import numpy as np
import pyninjotiff.ninjotiff as nt
import xarray as xr
from trollimage.xrimage import invert_scale_offset

from satpy.writers.core.image import ImageWriter

logger = logging.getLogger(__name__)


def convert_units(dataset, in_unit, out_unit):
    """Convert units of *dataset*.

    Convert dataset units for the benefit of writing NinJoTIFF.  The main
    background here is that NinJoTIFF would like brightness temperatures in °C,
    but satellite data files are in K.  For simplicity of implementation, this
    function can only convert from K to °C.

    This function will convert input data from K to °C and write the new unit
    in the ``"units"`` attribute.  When output and input units are equal, it
    returns the input dataset.

    Args:
        dataset (xarray.DataArray):
            Dataarray for which to convert the units.
        in_unit (str):
            Unit for input data.
        out_unit (str):
            Unit for output data.

    Returns:
        dataset, possibly with new units.
    """
    if in_unit == out_unit:
        return dataset

    if in_unit.lower() in {"k", "kelvin"} and out_unit.lower() in {"c", "celsius"}:
        logger.debug("Converting temperature units from K to °C")
        with xr.set_options(keep_attrs=True):
            new_dataset = dataset - 273.15
        new_dataset.attrs["units"] = out_unit
        return new_dataset

    # Other units not implemented.  Before Satpy 0.16.1 there was a
    # non-working implementation based on pint here.

    raise NotImplementedError(
            "NinJoTIFF unit conversion only implemented between K and C, not "
            f"between {in_unit!s} and {out_unit!s}")


class NinjoTIFFWriter(ImageWriter):
    """Writer for NinjoTiff files."""

    def __init__(self, tags=None, **kwargs):
        """Inititalize the writer."""
        ImageWriter.__init__(
            self, default_config_filename="writers/ninjotiff.yaml", **kwargs
        )

        self.tags = self.info.get("tags", None) if tags is None else tags
        if self.tags is None:
            self.tags = {}
        elif not isinstance(self.tags, dict):
            # if it's coming from a config file
            self.tags = dict(tuple(x.split("=")) for x in self.tags.split(","))

    def save_image(self, img, filename=None, compute=True, **kwargs):  # floating_point=False,
        """Save the image to the given *filename* in ninjotiff_ format.

        .. _ninjotiff: http://www.ssec.wisc.edu/~davidh/polar2grid/misc/NinJo_Satellite_Import_Formats.html
        """
        filename = filename or self.get_filename(**img.data.attrs)
        if img.mode.startswith("L") and (
            "ch_min_measurement_unit" not in kwargs
            or "ch_max_measurement_unit" not in kwargs
        ):
            try:
                scale, offset = img.get_scaling_from_history()
                scale, offset = invert_scale_offset(scale, offset)
            except ValueError as err:
                logger.warning(str(err))
            else:
                try:
                    # Here we know that the data if the image is scaled between 0 and 1
                    dmin = offset
                    dmax = scale + offset
                    ch_min_measurement_unit, ch_max_measurement_unit = np.minimum(dmin, dmax), np.maximum(dmin, dmax)
                    kwargs["ch_min_measurement_unit"] = ch_min_measurement_unit
                    kwargs["ch_max_measurement_unit"] = ch_max_measurement_unit
                except KeyError:
                    raise NotImplementedError(
                        "Don't know how to handle non-scale/offset-based enhancements yet."
                    )
        if img.mode.startswith("P"):
            img.data = img.data.astype(np.uint8)
        return [nt.save(img, filename, data_is_scaled_01=True, compute=compute, **kwargs)]

    def save_dataset(
        self, dataset, filename=None, fill_value=None, compute=True,
        convert_temperature_units=True, **kwargs
    ):
        """Save a dataset to ninjotiff format.

        This calls `save_image` in turn, but first preforms some unit conversion
        if necessary and desired.  Unit conversion can be suppressed by passing
        ``convert_temperature_units=False``.
        """
        nunits = kwargs.get("physic_unit", None)
        if nunits is None:
            try:
                options = nt.get_product_config(
                    kwargs["ninjo_product_name"], True, kwargs["ninjo_product_file"]
                )
                nunits = options["physic_unit"]
            except KeyError:
                pass
        if nunits is not None:
            try:
                units = dataset.attrs["units"]
            except KeyError:
                logger.warning(
                    "Saving to physical ninjo file without units defined in dataset!"
                )
            else:
                if convert_temperature_units:
                    dataset = convert_units(dataset, units, nunits)
                else:
                    logger.debug("Omitting unit conversion")
        return super(NinjoTIFFWriter, self).save_dataset(
            dataset, filename=filename, compute=compute, fill_value=fill_value, **kwargs
        )



"""Tests for the Insat3D reader."""

import datetime as dt
import os

import dask.array as da
import h5netcdf
import numpy as np
import pytest

from satpy import Scene
from satpy.readers.insat3d_img_l1b_h5 import (
    CHANNELS_BY_RESOLUTION,
    LUT_SUFFIXES,
    Insat3DIMGL1BH5FileHandler,
    get_lonlat_suffix,
    open_dataset,
    open_datatree,
)
from satpy.tests.utils import RANDOM_GEN, make_dataid

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path_factory

# real shape is 1, 11220, 11264
shape_1km = (1, 1122, 1126)
shape_4km = (1, 2816, 2805)
shape_8km = (1, 1408, 1402)
rad_units = "mW.cm-2.sr-1.micron-1"
alb_units = "%"
temp_units = "K"
chunks_1km = (1, 46, 1126)
values_1km = RANDOM_GEN.integers(0, 1000, shape_1km, dtype=np.uint16)
values_1km[0, 0, 0] = 0
values_4km = RANDOM_GEN.integers(0, 1000, shape_4km, dtype=np.uint16)
values_8km = RANDOM_GEN.integers(0, 1000, shape_8km, dtype=np.uint16)

values_by_resolution = {1000: values_1km,
                        4000: values_4km,
                        8000: values_8km}

lut_values_2 = np.arange(0, 1024 * 2, 2)
lut_values_3 = np.arange(0, 1024 * 3, 3)

dimensions = {"GeoX": shape_4km[2],
              "GeoY": shape_4km[1],
              "GeoX1": shape_8km[2],
              "GeoY1": shape_8km[1],
              "GeoX2": shape_1km[2],
              "GeoY2": shape_1km[1],
              "time": 1,
              "GreyCount": 1024,
              }
dimensions_by_resolution = {1000: ("GeoY2", "GeoX2"),
                            4000: ("GeoY", "GeoX"),
                            8000: ("GeoY1", "GeoX1")}

channel_names = {"vis": "Visible",
                 "mir": "Middle Infrared",
                 "swir": "Shortwave Infrared",
                 "tir1": "Thermal Infrared1",
                 "tir2": "Thermal Infrared2",
                 "wv": "Water Vapor"}

calibrated_names = {"": "Count",
                    "RADIANCE": "Radiance",
                    "ALBEDO": "Albedo",
                    "TEMP": "Brightness Temperature"}

calibrated_units = {"": "1",
                    "RADIANCE": "mW.cm-2.sr-1.micron-1",
                    "ALBEDO": "%",
                    "TEMP": "K"}

start_time = dt.datetime(2009, 6, 9, 9, 0)
end_time = dt.datetime(2009, 6, 9, 9, 30)
subsatellite_longitude = 82

time_pattern = "%d-%b-%YT%H:%M:%S"

global_attrs = {"Observed_Altitude(km)": 35778.490219,
                "Field_of_View(degrees)": 17.973925,
                "Acquisition_Start_Time": start_time.strftime(time_pattern),
                "Acquisition_End_Time": end_time.strftime(time_pattern),
                "Nominal_Central_Point_Coordinates(degrees)_Latitude_Longitude": [0.0, subsatellite_longitude],
                "Nominal_Altitude(km)": 36000.0,
                }


@pytest.fixture(scope="module")
def insat_filename(tmp_path_factory):
    """Create a fake insat 3d l1b file."""
    filename = tmp_path_factory.mktemp("data") / "3DIMG_25OCT2022_0400_L1B_STD_V01R00.h5"
    with h5netcdf.File(filename, mode="w") as h5f:
        h5f.dimensions = dimensions
        h5f.attrs.update(global_attrs)
        for resolution, channels in CHANNELS_BY_RESOLUTION.items():
            _create_channels(channels, h5f, resolution)
            _create_lonlats(h5f, resolution)

    yield filename
    import xarray as xr

    # clean xarray cache
    for key in xr.backends.file_manager.FILE_CACHE:
        if os.fspath(filename) in key[1]:
            ds = xr.backends.file_manager.FILE_CACHE.pop(key)
            ds.close()

    os.unlink(filename)


def mask_array(array):
    """Mask an array with nan instead of 0."""
    return np.where(array == 0, np.nan, array)


def _create_channels(channels, h5f, resolution):
    for channel in channels:
        var_name = "IMG_" + channel.upper()

        var = h5f.create_variable(var_name, ("time",) + dimensions_by_resolution[resolution], np.uint16,
                                  chunks=chunks_1km)
        var[:] = values_by_resolution[resolution]
        var.attrs["_FillValue"] = 0
        for suffix, lut_values in zip(LUT_SUFFIXES[channel], (lut_values_2, lut_values_3)):
            lut_name = "_".join((var_name, suffix))
            var = h5f.create_variable(lut_name, ("GreyCount",), float)
            var[:] = lut_values
            var.attrs["units"] = bytes(calibrated_units[suffix], "ascii")
            var.attrs["long_name"] = " ".join((channel_names[channel], calibrated_names[suffix]))


def _create_lonlats(h5f, resolution):
    lonlat_suffix = get_lonlat_suffix(resolution)
    for var_name in ["Longitude" + lonlat_suffix, "Latitude" + lonlat_suffix]:
        var = h5f.create_variable(var_name, dimensions_by_resolution[resolution], np.uint16,
                                  chunks=chunks_1km[1:])
        var[:] = values_by_resolution[resolution]
        var.attrs["scale_factor"] = 0.01
        var.attrs["add_offset"] = 0.0


def test_insat3d_backend_has_1km_channels(insat_filename):
    """Test the insat3d backend."""
    res = open_dataset(insat_filename, resolution=1000)
    assert res["IMG_VIS"].shape == shape_1km
    assert res["IMG_SWIR"].shape == shape_1km


@pytest.mark.parametrize(("resolution", "name", "shape", "expected_values", "expected_name", "expected_units"),
                         [(1000, "IMG_VIS_RADIANCE", shape_1km, mask_array(values_1km * 2),
                           "Visible Radiance", rad_units),
                          (1000, "IMG_VIS_ALBEDO", shape_1km, mask_array(values_1km * 3),
                           "Visible Albedo", alb_units),
                          (4000, "IMG_MIR_RADIANCE", shape_4km, mask_array(values_4km * 2),
                           "Middle Infrared Radiance", rad_units),
                          (4000, "IMG_MIR_TEMP", shape_4km, mask_array(values_4km * 3),
                           "Middle Infrared Brightness Temperature", temp_units),
                          (4000, "IMG_TIR1_RADIANCE", shape_4km, mask_array(values_4km * 2),
                           "Thermal Infrared1 Radiance", rad_units),
                          (4000, "IMG_TIR2_RADIANCE", shape_4km, mask_array(values_4km * 2),
                           "Thermal Infrared2 Radiance", rad_units),
                          (8000, "IMG_WV_RADIANCE", shape_8km, mask_array(values_8km * 2),
                           "Water Vapor Radiance", rad_units),
                          ])
def test_insat3d_has_calibrated_arrays(insat_filename,
                                       resolution, name, shape, expected_values, expected_name, expected_units):
    """Check that calibration happens as expected."""
    res = open_dataset(insat_filename, resolution=resolution)
    assert res[name].shape == shape
    np.testing.assert_allclose(res[name], expected_values)
    assert res[name].attrs["units"] == expected_units
    assert res[name].attrs["long_name"] == expected_name


def test_insat3d_has_dask_arrays(insat_filename):
    """Test that the backend uses dask."""
    res = open_dataset(insat_filename, resolution=1000)
    assert isinstance(res["IMG_VIS_RADIANCE"].data, da.Array)
    assert res["IMG_VIS"].chunks is not None


def test_insat3d_only_has_3_resolutions(insat_filename):
    """Test that we only accept 1000, 4000, 8000."""
    with pytest.raises(ValueError, match="Resolution 1024 not available. Available resolutions: 1000, 4000, 8000"):
        _ = open_dataset(insat_filename, resolution=1024)


@pytest.mark.parametrize("resolution", [1000, 4000, 8000, ])
def test_insat3d_returns_lonlat(insat_filename, resolution):
    """Test that lons and lats are loaded."""
    res = open_dataset(insat_filename, resolution=resolution)
    expected = values_by_resolution[resolution].squeeze() / 100.0
    assert isinstance(res["Latitude"].data, da.Array)
    np.testing.assert_allclose(res["Latitude"], expected)
    assert isinstance(res["Longitude"].data, da.Array)
    np.testing.assert_allclose(res["Longitude"], expected)


@pytest.mark.parametrize("resolution", [1000, 4000, 8000, ])
def test_insat3d_has_global_attributes(insat_filename, resolution):
    """Test that the backend supports global attributes."""
    res = open_dataset(insat_filename, resolution=resolution)
    assert res.attrs.keys() >= global_attrs.keys()


@pytest.mark.parametrize("resolution", [1000, 4000, 8000, ])
def test_insat3d_opens_datatree(insat_filename, resolution):
    """Test that a datatree is produced."""
    res = open_datatree(insat_filename)
    assert str(resolution) in res.keys()


def test_insat3d_datatree_has_global_attributes(insat_filename):
    """Test that the backend supports global attributes in the datatree."""
    res = open_datatree(insat_filename)
    assert res.attrs.keys() >= global_attrs.keys()


@pytest.mark.parametrize(("calibration", "expected_values"),
                         [("counts", values_1km),
                          ("radiance", mask_array(values_1km * 2)),
                          ("reflectance", mask_array(values_1km * 3))])
def test_filehandler_returns_data_array(insat_filehandler, calibration, expected_values):
    """Test that the filehandler can get dataarrays."""
    fh = insat_filehandler
    ds_info = None

    ds_id = make_dataid(name="VIS", resolution=1000, calibration=calibration)
    darr = fh.get_dataset(ds_id, ds_info)
    np.testing.assert_allclose(darr, expected_values.squeeze())
    assert darr.dims == ("y", "x")


def test_filehandler_returns_masked_data_in_space(insat_filehandler):
    """Test that the filehandler masks space pixels."""
    fh = insat_filehandler
    ds_info = None

    ds_id = make_dataid(name="VIS", resolution=1000, calibration="reflectance")
    darr = fh.get_dataset(ds_id, ds_info)
    assert np.isnan(darr[0, 0])


def test_insat3d_has_orbital_parameters(insat_filehandler):
    """Test that the filehandler returns data with orbital parameter attributes."""
    fh = insat_filehandler
    ds_info = None

    ds_id = make_dataid(name="VIS", resolution=1000, calibration="reflectance")
    darr = fh.get_dataset(ds_id, ds_info)

    assert "orbital_parameters" in darr.attrs
    assert "satellite_nominal_longitude" in darr.attrs["orbital_parameters"]
    assert darr.attrs["orbital_parameters"]["satellite_nominal_longitude"] == subsatellite_longitude
    assert "satellite_nominal_latitude" in darr.attrs["orbital_parameters"]
    assert "satellite_nominal_altitude" in darr.attrs["orbital_parameters"]
    assert "satellite_actual_altitude" in darr.attrs["orbital_parameters"]
    assert "platform_name" in darr.attrs
    assert "sensor" in darr.attrs


def test_filehandler_returns_coords(insat_filehandler):
    """Test that lon and lat can be loaded."""
    fh = insat_filehandler
    ds_info = None

    lon_id = make_dataid(name="longitude", resolution=1000)
    darr = fh.get_dataset(lon_id, ds_info)
    np.testing.assert_allclose(darr, values_1km.squeeze() / 100)


@pytest.fixture(scope="module")
def insat_filehandler(insat_filename):
    """Instantiate a Filehandler."""
    fileinfo = {}
    filetype = None
    fh = Insat3DIMGL1BH5FileHandler(insat_filename, fileinfo, filetype)
    return fh


def test_filehandler_returns_area(insat_filehandler):
    """Test that filehandle returns an area."""
    fh = insat_filehandler

    ds_id = make_dataid(name="MIR", resolution=4000, calibration="brightness_temperature")
    area_def = fh.get_area_def(ds_id)
    _ = area_def.get_lonlats(chunks=1000)
    assert subsatellite_longitude == area_def.crs.to_cf()["longitude_of_projection_origin"]
    np.testing.assert_allclose(area_def.area_extent, [-5620072.101427, -5640108.009097,
                                                      5620072.101427, 5644115.190631])


def test_filehandler_has_start_and_end_time(insat_filehandler):
    """Test that the filehandler handles start and end time."""
    fh = insat_filehandler

    assert fh.start_time == start_time
    assert fh.end_time == end_time


def test_satpy_load_array(insat_filename):
    """Test that satpy can load the VIS array."""
    scn = Scene(filenames=[os.fspath(insat_filename)], reader="insat3d_img_l1b_h5")
    scn.load(["VIS"])
    expected = mask_array(values_1km * 3).squeeze()
    np.testing.assert_allclose(scn["VIS"], expected)


def test_satpy_load_two_arrays(insat_filename):
    """Test that satpy can load the VIS array."""
    scn = Scene(filenames=[os.fspath(insat_filename)], reader="insat3d_img_l1b_h5")
    scn.load(["TIR1", "WV"])
    expected = mask_array(values_4km * 3).squeeze()
    np.testing.assert_allclose(scn["TIR1"], expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The HRIT base reader tests package."""

import bz2
import datetime as dt
import gzip
import os
from unittest import mock

import numpy as np
import pytest

from satpy.readers.core.hrit import HRITFileHandler
from satpy.tests.utils import RANDOM_GEN

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path


# From a compressed msg hrit file.
# uncompressed data field length 17223680
# compressed data field length 1578312
mda = {"file_type": 0, "total_header_length": 6198, "data_field_length": 17223680, "number_of_bits_per_pixel": 10,
       "number_of_columns": 3712, "number_of_lines": 464, "compression_flag_for_data": 0,
       "projection_name": b"GEOS(+000.0)                    ",
       "cfac": -13642337, "lfac": -13642337, "coff": 1856, "loff": 1856,
       "annotation_header": b"H-000-MSG4__-MSG4________-VIS006___-000001___-202208180730-C_",
       "cds_p_field": 64, "timestamp": (23605, 27911151), "GP_SC_ID": 324,
       "spectral_channel_id": 1,
       "segment_sequence_number": 1, "planned_start_segment_number": 1, "planned_end_segment_number": 8,
       "data_field_representation": 3,
       "image_segment_line_quality": np.array([(1, (0, 0), 1, 1, 0)] * 464,
                                              dtype=[("line_number_in_grid", ">i4"),
                                                     ("line_mean_acquisition", [("days", ">u2"),
                                                                                ("milliseconds", ">u4")]),
                                                     ("line_validity", "u1"),
                                                     ("line_radiometric_quality", "u1"),
                                                     ("line_geometric_quality", "u1")]),
       "projection_parameters": {"a": 6378169.0, "b": 6356583.8, "h": 35785831.0, "SSP_longitude": 0.0},
       "orbital_parameters": {}}

mda_compressed = mda.copy()
mda_compressed["data_field_length"] = 1578312
mda_compressed["compression_flag_for_data"] = 1


def new_get_hd(instance, hdr_info):
    """Generate some metadata."""
    if os.fspath(instance.filename).endswith(".C_"):
        m = mda_compressed.copy()
    else:
        m = mda.copy()
    return m


def new_get_hd_compressed(instance, hdr_info):
    """Generate some metadata."""
    m = mda.copy()
    m["compression_flag_for_data"] = 1
    m["data_field_length"] = 1578312
    return m


@pytest.fixture
def stub_hrit_file(tmp_path):
    """Create a stub hrit file."""
    filename = tmp_path / "some_hrit_file"
    create_stub_hrit(filename)
    return filename


def create_stub_hrit(filename, open_fun=open, meta=mda):
    """Create a stub hrit file."""
    stub_hrit_data = create_stub_hrit_data(meta)

    with open_fun(filename, mode="wb") as fd:
        fd.write(stub_hrit_data)
    return filename

def create_stub_hrit_data(meta):
    """Create the data for the stub hrit."""
    nbits = meta["number_of_bits_per_pixel"]
    lines = meta["number_of_lines"]
    cols = meta["number_of_columns"]
    total_bits = lines * cols * nbits
    arr = RANDOM_GEN.integers(0, 256,
                              size=int(total_bits / 8),
                              dtype=np.uint8)
    header_data = b" " * meta["total_header_length"]
    bytes_data = arr.tobytes()
    stub_hrit_data = header_data + bytes_data
    return stub_hrit_data


@pytest.fixture
def stub_bzipped_hrit_file(tmp_path):
    """Create a stub bzipped hrit file."""
    filename = tmp_path / "some_hrit_file.bz2"
    create_stub_hrit(filename, open_fun=bz2.open)
    return filename


@pytest.fixture
def stub_gzipped_hrit_file(tmp_path):
    """Create a stub gzipped hrit file."""
    filename = tmp_path / "some_hrit_file.gz"
    create_stub_hrit(filename, open_fun=gzip.open)
    return filename


@pytest.fixture
def stub_compressed_hrit_file(tmp_path):
    """Create a stub compressed hrit file."""
    filename = tmp_path / "some_hrit_file.C_"
    create_stub_hrit(filename, meta=mda_compressed)
    return filename


class TestHRITFileHandler:
    """Test the HRITFileHandler."""

    def setup_method(self, method):
        """Set up the hrit file handler for testing."""
        del method

        with mock.patch.object(HRITFileHandler, "_get_hd", new=new_get_hd):
            self.reader = HRITFileHandler("filename",
                                          {"platform_shortname": "MSG3",
                                           "start_time": dt.datetime(2016, 3, 3, 0, 0)},
                                          {"filetype": "info"},
                                          [mock.MagicMock(), mock.MagicMock(),
                                           mock.MagicMock()])

            self.reader.mda["cfac"] = 5
            self.reader.mda["lfac"] = 5
            self.reader.mda["coff"] = 10
            self.reader.mda["loff"] = 10
            self.reader.mda["projection_parameters"]["SSP_longitude"] = 44

    def test_get_xy_from_linecol(self):
        """Test get_xy_from_linecol."""
        x__, y__ = self.reader.get_xy_from_linecol(0, 0, (10, 10), (5, 5))
        assert -131072 == x__
        assert -131072 == y__
        x__, y__ = self.reader.get_xy_from_linecol(10, 10, (10, 10), (5, 5))
        assert x__ == 0
        assert y__ == 0
        x__, y__ = self.reader.get_xy_from_linecol(20, 20, (10, 10), (5, 5))
        assert 131072 == x__
        assert 131072 == y__

    def test_get_area_extent(self):
        """Test getting the area extent."""
        res = self.reader.get_area_extent((20, 20), (10, 10), (5, 5), 33)
        exp = (-71717.44995740513, -71717.44995740513,
               79266.655216079365, 79266.655216079365)
        assert res == exp

    def test_get_area_def(self):
        """Test getting an area definition."""
        from pyproj import CRS

        area = self.reader.get_area_def("VIS06")

        expected_crs = CRS(dict(proj="geos", a=6378169.0, b=6356583.8, h=35785831.0, lon_0=44.0, units="m"))
        assert area.crs == expected_crs

        assert area.area_extent == (-77771774058.38356, -77771774058.38356,
                                    30310525626438.438, 3720765401003.719)

    def test_read_band_filepath(self, stub_hrit_file):
        """Test reading a single band from a filepath."""
        self.reader.filename = stub_hrit_file

        res = self.reader.read_band("VIS006", None)
        assert res.compute().shape == (464, 3712)

    def test_read_band_FSFile(self, stub_hrit_file):
        """Test reading a single band from an FSFile."""
        import fsspec

        from satpy.readers.core.remote import FSFile

        filename = stub_hrit_file

        fs_file = fsspec.open(filename)
        self.reader.filename = FSFile(fs_file)

        res = self.reader.read_band("VIS006", None)
        assert res.compute().shape == (464, 3712)

    def test_read_band_bzipped2_filepath(self, stub_bzipped_hrit_file):
        """Test reading a single band from a bzipped file."""
        self.reader.filename = stub_bzipped_hrit_file

        res = self.reader.read_band("VIS006", None)
        assert res.compute().shape == (464, 3712)

    def test_read_band_gzip_stream(self, stub_gzipped_hrit_file):
        """Test reading a single band from a gzip stream."""
        import fsspec

        from satpy.readers.core.remote import FSFile

        filename = stub_gzipped_hrit_file

        fs_file = fsspec.open(filename, compression="gzip")
        self.reader.filename = FSFile(fs_file)

        res = self.reader.read_band("VIS006", None)
        assert res.compute().shape == (464, 3712)

    def test_start_end_time(self):
        """Test reading and converting start/end time."""
        assert self.reader.start_time == dt.datetime(2016, 3, 3, 0, 0)
        assert self.reader.start_time == self.reader.observation_start_time
        assert self.reader.end_time == dt.datetime(2016, 3, 3, 0, 0) + dt.timedelta(minutes=15)
        assert self.reader.end_time == self.reader.observation_end_time


def fake_decompress(filename):
    """Fake decompression."""
    del filename
    return create_stub_hrit_data(mda)


class TestHRITFileHandlerCompressed:
    """Test the HRITFileHandler with compressed segments."""

    def test_read_band_filepath(self, stub_compressed_hrit_file):
        """Test reading a single band from a filepath."""
        filename = stub_compressed_hrit_file

        with mock.patch("satpy.readers.core.hrit.decompress_buffer", side_effect=fake_decompress) as mock_decompress:
            with mock.patch.object(HRITFileHandler, "_get_hd", side_effect=new_get_hd, autospec=True) as get_hd:
                self.reader = HRITFileHandler(filename,
                                              {"platform_shortname": "MSG3",
                                               "start_time": dt.datetime(2016, 3, 3, 0, 0)},
                                              {"filetype": "info"},
                                              [mock.MagicMock(), mock.MagicMock(),
                                               mock.MagicMock()])

                res = self.reader.read_band("VIS006", None)
                assert get_hd.call_count == 1
                assert mock_decompress.call_count == 0
                assert res.compute().shape == (464, 3712)
                assert mock_decompress.call_count == 1



#!/usr/bin/python
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittests for generic image reader."""

import datetime as dt

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition
from rasterio.errors import NotGeoreferencedWarning

from satpy import Scene
from satpy.readers.generic_image import GenericImageFileHandler
from satpy.tests.utils import RANDOM_GEN, make_dataid

DATA_DATE = dt.datetime(2018, 1, 1)

X_SIZE = 100
Y_SIZE = 100
AREA_DEFINITION = AreaDefinition("geotiff_area", "ETRS89 / LAEA Europe", "ETRS89 / LAEA Europe",
                                 "EPSG:3035", X_SIZE, Y_SIZE,
                                 (2426378.0132, 1528101.2618, 6293974.6215, 5446513.5222))


@pytest.fixture
def random_image_channel():
    """Create random data."""
    return da.random.randint(0, 256, size=(Y_SIZE, X_SIZE), chunks=(50, 50)).astype(np.uint8)


random_image_channel_l = random_image_channel
random_image_channel_r = random_image_channel
random_image_channel_g = random_image_channel
random_image_channel_b = random_image_channel


@pytest.fixture
def alpha_channel():
    """Create alpha channel with fully transparent and opaque areas."""
    a__ = 255 * np.ones((Y_SIZE, X_SIZE), dtype=np.uint8)
    a__[:10, :10] = 0
    return da.from_array(a__, chunks=(50, 50))


@pytest.fixture
def random_image_channel_with_nans():
    """Create random data and replace a portion of it with NaN values."""
    arr = RANDOM_GEN.uniform(0., 1., size=(Y_SIZE, X_SIZE))
    arr[:10, :10] = np.nan
    return da.from_array(arr, chunks=(50, 50))


@pytest.fixture
def test_image_l(tmp_path, random_image_channel_l):
    """Create a test image with mode L."""
    dset = xr.DataArray(da.stack([random_image_channel_l]), dims=("bands", "y", "x"),
                        attrs={"name": "test_l", "start_time": DATA_DATE})
    dset["bands"] = ["L"]
    fname = tmp_path / "test_l.png"
    _save_image(dset, fname, "simple_image")

    return fname


@pytest.fixture
def test_image_l_nan(tmp_path, random_image_channel_with_nans):
    """Create a test image with mode L where data has NaN values."""
    dset = xr.DataArray(da.stack([random_image_channel_with_nans]), dims=("bands", "y", "x"),
                        attrs={"name": "test_l_nan", "start_time": DATA_DATE})
    dset["bands"] = ["L"]
    fname = tmp_path / "test_l_nan_nofillvalue.tif"
    _save_image(dset, fname, "geotiff")

    return fname


@pytest.fixture
def test_image_l_nan_fill_value(tmp_path, random_image_channel_with_nans):
    """Create a test image with mode L where data has NaN values and fill value is set."""
    dset = xr.DataArray(da.stack([random_image_channel_with_nans]), dims=("bands", "y", "x"),
                        attrs={"name": "test_l_nan", "start_time": DATA_DATE})
    dset["bands"] = ["L"]
    fname = tmp_path / "test_l_nan_fillvalue.tif"
    _save_image(dset, fname, "geotiff", fill_value=0)

    return fname


@pytest.fixture
def test_image_la(tmp_path, random_image_channel_l, alpha_channel):
    """Create a test image with mode LA."""
    dset = xr.DataArray(da.stack([random_image_channel_l, alpha_channel]),
                        dims=("bands", "y", "x"),
                        attrs={"name": "test_la", "start_time": DATA_DATE})
    dset["bands"] = ["L", "A"]
    fname = tmp_path / "20180101_0000_test_la.png"
    _save_image(dset, fname, "simple_image")

    return fname


@pytest.fixture
def test_image_rgb(tmp_path, random_image_channel_r, random_image_channel_g, random_image_channel_b):
    """Create a test image with mode RGB."""
    dset = xr.DataArray(da.stack([random_image_channel_r, random_image_channel_g, random_image_channel_b]),
                              dims=("bands", "y", "x"),
                              attrs={"name": "test_rgb",
                                     "start_time": DATA_DATE})
    dset["bands"] = ["R", "G", "B"]
    fname = tmp_path / "20180101_0000_test_rgb.tif"
    _save_image(dset, fname, "geotiff")

    return fname


@pytest.fixture
def rgba_dset(random_image_channel_r, random_image_channel_g, random_image_channel_b, alpha_channel):
    """Create an RGB dataset."""
    dset = xr.DataArray(
    da.stack([random_image_channel_r, random_image_channel_g, random_image_channel_b, alpha_channel]),
    dims=("bands", "y", "x"),
    attrs={"name": "test_rgba",
            "start_time": DATA_DATE})
    dset["bands"] = ["R", "G", "B", "A"]
    return dset


@pytest.fixture
def test_image_rgba(tmp_path, rgba_dset):
    """Create a test image with mode RGBA."""
    fname = tmp_path / "test_rgba.tif"
    _save_image(rgba_dset, fname, "geotiff")

    return fname


def _save_image(dset, fname, writer, fill_value=None):
    scn = Scene()
    scn["data"] = dset
    scn["data"].attrs["area"] = AREA_DEFINITION
    scn.save_dataset("data", str(fname), writer=writer, fill_value=fill_value)


def test_png_scene_l_mode(test_image_l):
    """Test reading a PNG image with L mode via satpy.Scene()."""
    with pytest.warns(NotGeoreferencedWarning, match=r"Dataset has no geotransform"):
        scn = Scene(reader="generic_image", filenames=[test_image_l])
    scn.load(["image"])
    _assert_image_common(scn, 1, None, None, np.float32)
    assert "area" not in scn["image"].attrs


def _assert_image_common(scn, channels, start_time, end_time, dtype):
    assert scn["image"].shape == (channels, Y_SIZE, X_SIZE)
    assert scn.sensor_names == {"images"}
    try:
        assert scn.start_time is start_time
        assert scn.end_time is end_time
    except AssertionError:
        assert scn.start_time == start_time
        assert scn.end_time == end_time
    assert scn["image"].dtype == dtype


def test_png_scene_la_mode(test_image_la):
    """Test reading a PNG image with LA mode via satpy.Scene()."""
    with pytest.warns(NotGeoreferencedWarning, match=r"Dataset has no geotransform"):
        scn = Scene(reader="generic_image", filenames=[test_image_la])
    scn.load(["image"])
    data = da.compute(scn["image"].data)
    assert np.sum(np.isnan(data)) == 100
    assert "area" not in scn["image"].attrs
    _assert_image_common(scn, 1, DATA_DATE, DATA_DATE, np.float32)


def test_geotiff_scene_rgb(test_image_rgb):
    """Test reading geotiff image in RGB mode via satpy.Scene()."""
    scn = Scene(reader="generic_image", filenames=[test_image_rgb])
    scn.load(["image"])
    assert scn["image"].area == AREA_DEFINITION
    _assert_image_common(scn, 3, DATA_DATE, DATA_DATE, np.float32)


def test_geotiff_scene_rgba(test_image_rgba):
    """Test reading geotiff image in RGBA mode via satpy.Scene()."""
    scn = Scene(reader="generic_image", filenames=[test_image_rgba])
    scn.load(["image"])
    _assert_image_common(scn, 3, None, None, np.float32)
    assert scn["image"].area == AREA_DEFINITION


def test_geotiff_scene_nan_fill_value(test_image_l_nan_fill_value):
    """Test reading geotiff image with fill value set via satpy.Scene()."""
    scn = Scene(reader="generic_image", filenames=[test_image_l_nan_fill_value])
    scn.load(["image"])
    assert np.sum(scn["image"].data[0][:10, :10].compute()) == 0
    _assert_image_common(scn, 1, None, None, np.uint8)

def test_geotiff_scene_nan(test_image_l_nan):
    """Test reading geotiff image with NaN values in it via satpy.Scene()."""
    scn = Scene(reader="generic_image", filenames=[test_image_l_nan])
    scn.load(["image"])
    assert np.all(np.isnan(scn["image"].data[0][:10, :10].compute()))
    _assert_image_common(scn, 1, None, None, np.float32)


def test_GenericImageFileHandler(test_image_rgba):
    """Test direct use of the reader."""
    from satpy.readers.generic_image import GenericImageFileHandler

    fname_info = {"start_time": DATA_DATE}
    ftype_info = {}
    reader = GenericImageFileHandler(test_image_rgba, fname_info, ftype_info)

    data_id = make_dataid(name="image")
    assert reader.file_content
    assert reader.finfo["filename"] == test_image_rgba
    assert reader.finfo["start_time"] == DATA_DATE
    assert reader.finfo["end_time"] == DATA_DATE
    assert reader.area == AREA_DEFINITION
    assert reader.get_area_def(None) == AREA_DEFINITION
    assert reader.start_time == DATA_DATE
    assert reader.end_time == DATA_DATE

    dataset = reader.get_dataset(data_id, {})
    assert isinstance(dataset, xr.DataArray)
    assert "spatial_ref" in dataset.coords
    assert np.all(np.isnan(dataset.data[:, :10, :10].compute()))


class FakeGenericImageFileHandler(GenericImageFileHandler):
    """Fake file handler."""

    def __init__(self, filename, filename_info, filetype_info, file_content, **kwargs):
        """Get fake file content from 'get_test_content'."""
        super(GenericImageFileHandler, self).__init__(filename, filename_info, filetype_info)
        self.file_content = file_content
        self.dataset_name = None
        self.file_content.update(kwargs)


def test_GenericImageFileHandler_no_masking_for_float(rgba_dset):
    """Test direct use of the reader for float_data."""
    # do nothing if not integer
    float_data = rgba_dset / 255.
    reader = FakeGenericImageFileHandler("dummy", {}, {}, {"image": float_data})
    assert reader.get_dataset(make_dataid(name="image"), {}) is float_data


def test_GenericImageFileHandler_masking_for_integer(rgba_dset):
    """Test direct use of the reader for float_data."""
    # masking if integer
    data = rgba_dset.astype(np.uint32)
    assert data.bands.size == 4
    reader = FakeGenericImageFileHandler("dummy", {}, {}, {"image": data})
    ret_data = reader.get_dataset(make_dataid(name="image"), {})
    assert ret_data.bands.size == 3


def test_GenericImageFileHandler_datasetid(test_image_rgba):
    """Test direct use of the reader."""
    fname_info = {"start_time": DATA_DATE}
    ftype_info = {}
    reader = GenericImageFileHandler(test_image_rgba, fname_info, ftype_info)

    data_id = make_dataid(name="image-custom")
    assert reader.file_content
    dataset = reader.get_dataset(data_id, {})
    assert isinstance(dataset, xr.DataArray)


@pytest.fixture
def reader_l_nan_fill_value(test_image_l_nan_fill_value):
    """Create GenericImageFileHandler."""
    fname_info = {"start_time": DATA_DATE}
    ftype_info = {}
    return GenericImageFileHandler(test_image_l_nan_fill_value, fname_info, ftype_info)


def test_GenericImageFileHandler_nodata_nan_mask(reader_l_nan_fill_value):
    """Test nodata handling with direct use of the reader with nodata handling: nan_mask."""
    data_id = make_dataid(name="image-custom")
    assert reader_l_nan_fill_value.file_content
    info = {"nodata_handling": "nan_mask"}
    dataset = reader_l_nan_fill_value.get_dataset(data_id, info)
    assert isinstance(dataset, xr.DataArray)
    assert np.all(np.isnan(dataset.data[0][:10, :10].compute()))
    assert np.isnan(dataset.attrs["_FillValue"])


def test_GenericImageFileHandler_nodata_fill_value(reader_l_nan_fill_value):
    """Test nodata handling with direct use of the reader with nodata handling: fill_value."""
    info = {"nodata_handling": "fill_value"}
    data_id = make_dataid(name="image-custom")
    dataset = reader_l_nan_fill_value.get_dataset(data_id, info)
    assert isinstance(dataset, xr.DataArray)
    assert np.sum(dataset.data[0][:10, :10].compute()) == 0
    assert dataset.attrs["_FillValue"] == 0


def test_GenericImageFileHandler_nodata_nan_mask_default(reader_l_nan_fill_value):
    """Test nodata handling with direct use of the reader with default nodata handling."""
    data_id = make_dataid(name="image-custom")
    dataset = reader_l_nan_fill_value.get_dataset(data_id, {})
    assert isinstance(dataset, xr.DataArray)
    assert np.sum(dataset.data[0][:10, :10].compute()) == 0
    assert dataset.attrs["_FillValue"] == 0



# Copyright (c) 2019-2023 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for both SEVIRI and FCI L2 products in GRIB2 format.

References:
    FM 92 GRIB Edition 2
    https://www.wmo.int/pages/prog/www/WMOCodes/Guides/GRIB/GRIB2_062006.pdf
    EUMETSAT Product Navigator
    https://navigator.eumetsat.int/
"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import xarray as xr

from satpy.readers.core._geos_area import get_area_definition, get_geos_area_naming
from satpy.readers.core.eum import get_service_mode
from satpy.readers.core.fci import calculate_area_extent as fci_calculate_area_extent
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.seviri import PLATFORM_DICT as SEVIRI_PLATFORM_DICT
from satpy.readers.core.seviri import REPEAT_CYCLE_DURATION as SEVIRI_REPEAT_CYCLE_DURATION
from satpy.readers.core.seviri import REPEAT_CYCLE_DURATION_RSS as SEVIRI_REPEAT_CYCLE_DURATION_RSS
from satpy.readers.core.seviri import calculate_area_extent as seviri_calculate_area_extent
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()

try:
    import eccodes as ec
except ImportError:
    raise ImportError(
                "Missing eccodes-python and/or eccodes C-library installation. Use conda to install eccodes")

logger = logging.getLogger(__name__)


class EUML2GribFileHandler(BaseFileHandler):
    """Reader class for EUM L2 products in GRIB format."""

    calculate_area_extent = None

    def __init__(self, filename, filename_info, filetype_info):
        """Read the global attributes and prepare for dataset reading."""
        super().__init__(filename, filename_info, filetype_info)
        # Turn on support for multiple fields in single GRIB messages (required for SEVIRI L2 files)
        ec.codes_grib_multi_support_on()

        if "seviri" in self.filetype_info["file_type"]:
            self.sensor = "seviri"
            self.PLATFORM_NAME = SEVIRI_PLATFORM_DICT[self.filename_info["spacecraft"]]
        elif "fci" in self.filetype_info["file_type"]:
            self.sensor = "fci"
            self.PLATFORM_NAME = f"MTG-i{self.filename_info['spacecraft_id']}"
        pass

    @property
    def start_time(self):
        """Return the sensing start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Return the sensing end time."""
        if self.sensor == "seviri":
            try:
                delta = SEVIRI_REPEAT_CYCLE_DURATION_RSS if self._ssp_lon == 9.5 else SEVIRI_REPEAT_CYCLE_DURATION
                return self.start_time + dt.timedelta(minutes=delta)
            except AttributeError:
                # If dataset and metadata (ssp_lon) have not yet been loaded, return None
                return None
        elif self.sensor == "fci":
            return self.filename_info["end_time"]

    def get_area_def(self, dataset_id):
        """Return the area definition for a dataset."""
        # Compute the dictionary with the area extension

        self._area_dict["column_step"] = dataset_id["resolution"]
        self._area_dict["line_step"] = dataset_id["resolution"]

        if self.sensor == "seviri":
            area_extent = seviri_calculate_area_extent(self._area_dict)

        elif self.sensor == "fci":
            area_extent = fci_calculate_area_extent(self._area_dict)

        # Call the get_area_definition function to obtain the area
        area_def = get_area_definition(self._pdict, area_extent)

        return area_def

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using the parameter_number key in dataset_info.

        In a previous version of the reader, the attributes (nrows, ncols, ssp_lon) and projection information
        (pdict and area_dict) were computed while initializing the file handler. Also the code would break out from
        the While-loop below as soon as the correct parameter_number was found. This has now been revised becasue the
        reader would sometimes give corrupt information about the number of messages in the file and the dataset
        dimensions within a given message if the file was only partly read (not looping over all messages) in an earlier
        instance.
        """
        logger.debug("Reading in file to get dataset with parameter number %d.",
                     dataset_info["parameter_number"])

        xarr = None
        message_found = False
        with open(self.filename, "rb") as fh:

            # Iterate over all messages and fetch data when the correct parameter number is found
            while True:
                gid = ec.codes_grib_new_from_file(fh)

                if gid is None:
                    if not message_found:
                        # Could not obtain a valid message ID from the grib file
                        logger.warning("Could not find parameter_number %d in GRIB file, no valid Dataset created",
                                       dataset_info["parameter_number"])
                    break

                # Check if the parameter number in the GRIB message corresponds to the required key
                parameter_number = self._get_from_msg(gid, "parameterNumber")

                if parameter_number == dataset_info["parameter_number"]:

                    self._res = dataset_id["resolution"]
                    self._read_attributes(gid)

                    # Read the missing value
                    missing_value = self._get_from_msg(gid, "missingValue")

                    # Retrieve values and metadata from the GRIB message, masking the values equal to missing_value
                    xarr = self._get_xarray_from_msg(gid)

                    xarr.data = da.where(xarr.data == missing_value, np.nan, xarr.data)

                    ec.codes_release(gid)

                    # Combine all metadata into the dataset attributes and break out of the loop
                    xarr.attrs.update(dataset_info)
                    xarr.attrs.update(self._get_attributes())

                    message_found = True

                else:
                    # The parameter number is not the correct one, release gid and skip to next message
                    ec.codes_release(gid)

        return xarr

    def _read_attributes(self, gid):
        """Read the parameter attributes from the message and create the projection and area dictionaries."""
        # Read SSP and date/time
        self._ssp_lon = self._get_from_msg(gid, "longitudeOfSubSatellitePointInDegrees")

        # Read number of points on the x and y axes
        self._nrows = self._get_from_msg(gid, "Ny")
        self._ncols = self._get_from_msg(gid, "Nx")

        # Creates the projection and area dictionaries
        self._pdict, self._area_dict = self._get_proj_area(gid)

    def _get_proj_area(self, gid):
        """Compute the dictionary with the projection and area definition from a GRIB message.

        Args:
            gid: The ID of the GRIB message.

        Returns:
            tuple: A tuple of two dictionaries for the projection and the area definition.
                pdict:
                    a: Earth major axis [m]
                    b: Earth minor axis [m]
                    h: Height over surface [m]
                    ssp_lon: longitude of subsatellite point [deg]
                    nlines: number of lines
                    ncols: number of columns
                    a_name: name of the area
                    a_desc: description of the area
                    p_id: id of the projection
                area_dict:
                    center_point: coordinate of the center point
                    north: coodinate of the north limit
                    east: coodinate of the east limit
                    west: coodinate of the west limit
                    south: coodinate of the south limit
        """
        # Get name of area definition
        area_naming_input_dict = {"platform_name": "msg",
                                  "instrument_name": self.sensor,
                                  "resolution": self._res,
                                  }

        area_naming = get_geos_area_naming({**area_naming_input_dict,
                                            **get_service_mode(self.sensor, self._ssp_lon)})

        # Read all projection and area parameters from the message
        earth_major_axis_in_meters = self._get_from_msg(gid, "earthMajorAxis") * 1000.0  # [m]
        earth_minor_axis_in_meters = self._get_from_msg(gid, "earthMinorAxis") * 1000.0  # [m]

        if self.sensor == "seviri":
            earth_major_axis_in_meters = self._scale_earth_axis(earth_major_axis_in_meters)
            earth_minor_axis_in_meters = self._scale_earth_axis(earth_minor_axis_in_meters)

        nr_in_radius_of_earth = self._get_from_msg(gid, "NrInRadiusOfEarth")
        xp_in_grid_lengths = self._get_from_msg(gid, "XpInGridLengths")
        h_in_meters = earth_major_axis_in_meters * (nr_in_radius_of_earth - 1.0)  # [m]

        # Create the dictionary with the projection data
        pdict = {
            "a": earth_major_axis_in_meters,
            "b": earth_minor_axis_in_meters,
            "h": h_in_meters,
            "ssp_lon": self._ssp_lon,
            "nlines": self._ncols,
            "ncols": self._nrows,
            "a_name": area_naming["area_id"],
            "a_desc": area_naming["description"],
            "p_id": "",
        }

        if self.sensor == "seviri":
            # Compute the dictionary with the area extension
            area_dict = {
                "center_point": xp_in_grid_lengths,
                "north": self._nrows,
                "east": 1,
                "west": self._ncols,
                "south": 1,
            }

        elif self.sensor == "fci":
            area_dict = {
                "nlines": self._ncols,
                "ncols": self._nrows,
            }

        return pdict, area_dict

    @staticmethod
    def _scale_earth_axis(data):
        """Scale Earth axis data to make sure the value matched the expected unit [m].

        The earthMinorAxis value stored in the MPEF aerosol over sea product prior to December 12, 2022 has the wrong
        unit and this method provides a flexible work-around by making sure that all earth axis values are scaled such
        that they are on the order of millions of meters as expected by the reader.

        """
        scale_factor = 10 ** np.ceil(np.log10(1e6/data))
        return data * scale_factor

    def _get_xarray_from_msg(self, gid):
        """Read the values from the GRIB message and return a DataArray object.

        Args:
            gid: The ID of the GRIB message.

        Returns:
            The DataArray containing the retrieved values.
        """
        # Data from GRIB message are read into an Xarray...
        xarr = xr.DataArray(da.from_array(ec.codes_get_values(
            gid).reshape(self._nrows, self._ncols), CHUNK_SIZE), dims=("y", "x"))

        return xarr

    def _get_attributes(self):
        """Create a dictionary of  attributes to be added to the dataset.

        Returns:
            dict: A dictionary of parameter attributes.
                ssp_lon: longitude of subsatellite point
                sensor: name of sensor
                platform_name: name of the platform
        """
        orbital_parameters = {
            "projection_longitude": self._ssp_lon
        }

        attributes = {"orbital_parameters": orbital_parameters, "sensor": self.sensor,
                      "platform_name": self.PLATFORM_NAME}

        return attributes

    @staticmethod
    def _get_from_msg(gid, key):
        """Get a value from the GRIB message based on the key, return None if missing.

        Args:
            gid: The ID of the GRIB message.
            key: The key of the required attribute.

        Returns:
            The retrieved attribute or None if the key is missing.
        """
        try:
            attr = ec.codes_get(gid, key)
        except ec.KeyValueNotFoundError:
            logger.warning("Key %s not found in GRIB message", key)
            attr = None
        return attr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Compositors for cloud products."""

import numpy as np

from satpy.composites.core import GenericCompositor, SingleBandCompositor


class CloudCompositorWithoutCloudfree(SingleBandCompositor):
    """Put cloud-free pixels as fill_value_color in palette."""

    def __call__(self, projectables, **info):
        """Create the composite."""
        if len(projectables) != 2:
            raise ValueError("Expected 2 datasets, got %d" %
                             (len(projectables), ))
        data, status = projectables
        valid = status != status.attrs["_FillValue"]
        status_cloud_free = status % 2 == 1  # bit 0 is set
        cloud_free = np.logical_and(valid, status_cloud_free)
        if "bad_optical_conditions" in status.attrs.get("flag_meanings", "") and data.name == "cmic_cre":
            bad_optical_conditions = np.bitwise_and(np.right_shift(status, 1), 1)
            cloud_free = np.logical_and(cloud_free, np.logical_not(bad_optical_conditions))
        # Where condition is true keep data, in other place update to scaled_FillValue:
        data = data.where(np.logical_not(cloud_free), data.attrs["scaled_FillValue"])
        # Update not cloudfree product and nodata to NaN (already done for scaled vars in the reader)
        # Keep cloudfree or valid product
        data = data.where(np.logical_or(cloud_free, data != data.attrs["scaled_FillValue"]), np.nan)
        res = SingleBandCompositor.__call__(self, [data], **data.attrs)
        res.attrs["_FillValue"] = np.nan
        return res


class CloudCompositorCommonMask(SingleBandCompositor):
    """Put cloud-free pixels as fill_value_color in palette."""

    def __call__(self, projectables, **info):
        """Create the composite."""
        if len(projectables) != 2:
            raise ValueError("Expected 2 datasets, got %d" %
                             (len(projectables), ))
        data, cma = projectables
        valid_cma = cma != cma.attrs["_FillValue"]
        valid_prod = data != data.attrs["_FillValue"]
        valid_prod = np.logical_and(valid_prod, np.logical_not(np.isnan(data)))
        # Update valid_cma and not valid_prod means: keep not valid cma or valid prod
        data = data.where(np.logical_or(np.logical_not(valid_cma), valid_prod),
                          data.attrs["scaled_FillValue"])
        data = data.where(np.logical_or(valid_prod, valid_cma), np.nan)
        res = SingleBandCompositor.__call__(self, [data], **data.attrs)
        res.attrs["_FillValue"] = np.nan
        return res


class PrecipCloudsRGB(GenericCompositor):
    """Precipitation clouds compositor."""

    def __call__(self, projectables, *args, **kwargs):
        """Make an RGB image out of the three probability categories of the NWCSAF precip product."""
        projectables = self.match_data_arrays(projectables)
        light = projectables[0]
        moderate = projectables[1]
        intense = projectables[2]
        status_flag = projectables[3]

        if np.bitwise_and(status_flag, 4).any():
            # AMSU is used
            maxs1 = 70
            maxs2 = 70
            maxs3 = 100
        else:
            # avhrr only
            maxs1 = 30
            maxs2 = 50
            maxs3 = 40

        scalef3 = 1.0 / maxs3 - 1 / 255.0
        scalef2 = 1.0 / maxs2 - 1 / 255.0
        scalef1 = 1.0 / maxs1 - 1 / 255.0

        p1data = (light*scalef1).where(light != 0)
        p1data = p1data.where(light != light.attrs["_FillValue"])
        p1data.attrs = light.attrs
        data = moderate*scalef2
        p2data = data.where(moderate != 0)
        p2data = p2data.where(moderate != moderate.attrs["_FillValue"])
        p2data.attrs = moderate.attrs
        data = intense*scalef3
        p3data = data.where(intense != 0)
        p3data = p3data.where(intense != intense.attrs["_FillValue"])
        p3data.attrs = intense.attrs

        res = super(PrecipCloudsRGB, self).__call__((p3data, p2data, p1data),
                                                    *args, **kwargs)
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Advanced Himawari Imager (AHI) standard format data reader.

References:
    - Himawari-8/9 Himawari Standard Data User's Guide
    - http://www.data.jma.go.jp/mscweb/en/himawari89/space_segment/spsg_ahi.html

Time Information
****************

AHI observations use the idea of a "nominal" time and an "observation" time.
The "nominal" time or repeat cycle is the overall window when the instrument
can record data, usually at a specific and consistent interval. The
"observation" time is when the data was actually observed inside the nominal
window. These two times are stored in a sub-dictionary in the metadata calls
``time_parameters``. Nominal time can be accessed from the
``nominal_start_time`` and ``nominal_end_time`` metadata keys and
observation time from the ``observation_start_time`` and
``observation_end_time`` keys. Observation time can also be accessed from the
parent (``.attrs``) dictionary as the ``start_time`` and ``end_time`` keys.

Satellite Position
******************

As discussed in the :ref:`orbital_parameters` documentation, a satellite
position can be described by a specific "actual" position, a "nominal"
position, a "projection" position, or sometimes a "nadir" position. Not all
readers are able to produce all of these positions. In the case of AHI HSD data
we have an "actual" and "projection" position. For a lot of sensors/readers
though, the "actual" position values do not change between bands or segments
of the same time step (repeat cycle). AHI HSD files contain varying values for
the actual position.

Other components in Satpy use this actual satellite
position to generate other values (ex. sensor zenith angles). If these values
are not consistent between bands then Satpy (dask) will not be able to share
these calculations (generate one sensor zenith angle for band 1, another for
band 2, etc) even though there is rarely a noticeable difference. To deal with
this this reader has an option ``round_actual_position`` that defaults to
``True`` and will round the "actual" position (longitude, latitude, altitude)
in a way to produce as consistent a position between bands as possible.

"""

import datetime as dt
import logging
import os
import warnings
import weakref

import dask.array as da
import numpy as np
import xarray as xr

from satpy._compat import cached_property
from satpy.readers.core._geos_area import get_area_definition, get_area_extent
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.utils import (
    apply_rad_correction,
    get_earth_radius,
    get_geostationary_mask,
    get_user_calibration_factors,
    np2str,
    unzip_file,
)
from satpy.utils import normalize_low_res_chunks

AHI_CHANNEL_NAMES = ("1", "2", "3", "4", "5",
                     "6", "7", "8", "9", "10",
                     "11", "12", "13", "14", "15", "16")

logger = logging.getLogger("ahi_hsd")

# Basic information block:
_BASIC_INFO_TYPE = np.dtype([("hblock_number", "u1"),
                             ("blocklength", "<u2"),
                             ("total_number_of_hblocks", "<u2"),
                             ("byte_order", "u1"),
                             ("satellite", "S16"),
                             ("proc_center_name", "S16"),
                             ("observation_area", "S4"),
                             ("other_observation_info", "S2"),
                             ("observation_timeline", "<u2"),
                             ("observation_start_time", "f8"),
                             ("observation_end_time", "f8"),
                             ("file_creation_time", "f8"),
                             ("total_header_length", "<u4"),
                             ("total_data_length", "<u4"),
                             ("quality_flag1", "u1"),
                             ("quality_flag2", "u1"),
                             ("quality_flag3", "u1"),
                             ("quality_flag4", "u1"),
                             ("file_format_version", "S32"),
                             ("file_name", "S128"),
                             ("spare", "S40"),
                             ])

# Data information block
_DATA_INFO_TYPE = np.dtype([("hblock_number", "u1"),
                            ("blocklength", "<u2"),
                            ("number_of_bits_per_pixel", "<u2"),
                            ("number_of_columns", "<u2"),
                            ("number_of_lines", "<u2"),
                            ("compression_flag_for_data", "u1"),
                            ("spare", "S40"),
                            ])

# Projection information block
# See footnote 2; LRIT/HRIT Global Specification Section 4.4, CGMS, 1999)
_PROJ_INFO_TYPE = np.dtype([("hblock_number", "u1"),
                            ("blocklength", "<u2"),
                            ("sub_lon", "f8"),
                            ("CFAC", "<u4"),
                            ("LFAC", "<u4"),
                            ("COFF", "f4"),
                            ("LOFF", "f4"),
                            ("distance_from_earth_center", "f8"),
                            ("earth_equatorial_radius", "f8"),
                            ("earth_polar_radius", "f8"),
                            ("req2_rpol2_req2", "f8"),
                            ("rpol2_req2", "f8"),
                            ("req2_rpol2", "f8"),
                            ("coeff_for_sd", "f8"),
                            # Note: processing center use only:
                            ("resampling_types", "<i2"),
                            # Note: processing center use only:
                            ("resampling_size", "<i2"),
                            ("spare", "S40"),
                            ])

# Navigation information block
_NAV_INFO_TYPE = np.dtype([("hblock_number", "u1"),
                           ("blocklength", "<u2"),
                           ("navigation_info_time", "f8"),
                           ("SSP_longitude", "f8"),
                           ("SSP_latitude", "f8"),
                           ("distance_earth_center_to_satellite", "f8"),
                           ("nadir_longitude", "f8"),
                           ("nadir_latitude", "f8"),
                           ("sun_position", "f8", (3,)),
                           ("moon_position", "f8", (3,)),
                           ("spare", "S40"),
                           ])

# Calibration information block
_CAL_INFO_TYPE = np.dtype([("hblock_number", "u1"),
                           ("blocklength", "<u2"),
                           ("band_number", "<u2"),
                           ("central_wave_length", "f8"),
                           ("valid_number_of_bits_per_pixel", "<u2"),
                           ("count_value_error_pixels", "<u2"),
                           ("count_value_outside_scan_pixels", "<u2"),
                           ("gain_count2rad_conversion", "f8"),
                           ("offset_count2rad_conversion", "f8"),
                           ])

# Infrared band (Band No. 7 – 16)
# (Band No. 2 – 5: backup operation (See Table 4 bb))
_IRCAL_INFO_TYPE = np.dtype([("c0_rad2tb_conversion", "f8"),
                             ("c1_rad2tb_conversion", "f8"),
                             ("c2_rad2tb_conversion", "f8"),
                             ("c0_tb2rad_conversion", "f8"),
                             ("c1_tb2rad_conversion", "f8"),
                             ("c2_tb2rad_conversion", "f8"),
                             ("speed_of_light", "f8"),
                             ("planck_constant", "f8"),
                             ("boltzmann_constant", "f8"),
                             ("spare", "S40"),
                             ])

# Visible, near-infrared band (Band No. 1 – 6)
# (Band No. 1: backup operation (See Table 4 bb))
_VISCAL_INFO_TYPE = np.dtype([("coeff_rad2albedo_conversion", "f8"),
                              ("coeff_update_time", "f8"),
                              ("cali_gain_count2rad_conversion", "f8"),
                              ("cali_offset_count2rad_conversion", "f8"),
                              ("spare", "S80"),
                              ])

# 6 Inter-calibration information block
_INTER_CALIBRATION_INFO_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u2"),
    ("gsics_calibration_intercept", "f8"),
    ("gsics_calibration_slope", "f8"),
    ("gsics_calibration_coeff_quadratic_term", "f8"),
    ("gsics_std_scn_radiance_bias", "f8"),
    ("gsics_std_scn_radiance_bias_uncertainty", "f8"),
    ("gsics_std_scn_radiance", "f8"),
    ("gsics_correction_starttime", "f8"),
    ("gsics_correction_endtime", "f8"),
    ("gsics_radiance_validity_upper_lim", "f4"),
    ("gsics_radiance_validity_lower_lim", "f4"),
    ("gsics_filename", "S128"),
    ("spare", "S56"),
])

# 7 Segment information block
_SEGMENT_INFO_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u2"),
    ("total_number_of_segments", "u1"),
    ("segment_sequence_number", "u1"),
    ("first_line_number_of_image_segment", "u2"),
    ("spare", "S40"),
])

# 8 Navigation correction information block
_NAVIGATION_CORRECTION_INFO_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u2"),
    ("center_column_of_rotation", "f4"),
    ("center_line_of_rotation", "f4"),
    ("amount_of_rotational_correction", "f8"),
    ("numof_correction_info_data", "<u2"),
])

# Navigation correction sub-info
_NAVIGATION_CORRECTION_SUBINFO_TYPE = np.dtype([
    ("line_number_after_rotation", "<u2"),
    ("shift_amount_for_column_direction", "f4"),
    ("shift_amount_for_line_direction", "f4"),
])

# 9 Observation time information block
_OBSERVATION_TIME_INFO_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u2"),
    ("number_of_observation_times", "<u2"),
])

_OBSERVATION_LINE_TIME_INFO_TYPE = np.dtype([
    ("line_number", "<u2"),
    ("observation_time", "f8"),
])

# 10 Error information block
_ERROR_INFO_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u4"),
    ("number_of_error_info_data", "<u2"),
])

_ERROR_LINE_INFO_TYPE = np.dtype([
    ("line_number", "<u2"),
    ("numof_error_pixels_per_line", "<u2"),
])

# 11 Spare block
_SPARE_TYPE = np.dtype([
    ("hblock_number", "u1"),
    ("blocklength", "<u2"),
    ("spare", "S256")
])


class AHIHSDFileHandler(BaseFileHandler):
    """AHI standard format reader.

    The AHI sensor produces data for some pixels outside the Earth disk (i,e:
    atmospheric limb or deep space pixels).
    By default, these pixels are masked out as they contain data of limited
    or no value, but some applications do require these pixels.
    It is therefore possible to override the default behaviour and perform no
    masking of non-Earth pixels.

    In order to change the default behaviour, use the 'mask_space' variable
    as part of ``reader_kwargs`` upon Scene creation::

        import satpy
        import glob

        filenames = glob.glob('*FLDK*.dat')
        scene = satpy.Scene(filenames,
                            reader='ahi_hsd',
                            reader_kwargs={'mask_space': False})
        scene.load([0.6])

    The AHI HSD data files contain multiple VIS channel calibration
    coefficients. By default, the updated coefficients in header block 6
    are used. If the user prefers the default calibration coefficients from
    block 5 then they can pass calib_mode='nominal' when creating a scene::

        import satpy
        import glob

        filenames = glob.glob('*FLDK*.dat')
        scene = satpy.Scene(filenames,
                            reader='ahi_hsd',
                            reader_kwargs={'calib_mode': 'update'})
        scene.load([0.6])

    Alternative AHI calibrations are also available, such as GSICS
    coefficients. As such, you can supply custom per-channel correction
    by setting calib_mode='custom' and passing correction factors via::

        user_calibration={'chan': ['slope': slope, 'offset': offset]}

    Where slo and off are per-channel slope and offset coefficients defined by::

        rad_leo = (rad_geo - off) / slo

    If you do not have coefficients for a particular band, then by default the
    slope will be set to 1 .and the offset to 0.::

        import satpy
        import glob

        # Load bands 7, 14 and 15, but we only have coefs for 7+14
        calib_dict = {'B07': {'slope': 0.99, 'offset': 0.002},
                      'B14': {'slope': 1.02, 'offset': -0.18}}

        filenames = glob.glob('*FLDK*.dat')
        scene = satpy.Scene(filenames,
                            reader='ahi_hsd',
                            reader_kwargs={'user_calibration': calib_dict)
        # B15 will not have custom radiance correction applied.
        scene.load(['B07', 'B14', 'B15'])

    By default, user-supplied calibrations / corrections are applied to the
    radiance data in accordance with the GSICS standard defined in the
    equation above. However, user-supplied gain and offset values for
    converting digital number into radiance via Rad = DN * gain + offset are
    also possible. To supply your own factors, supply a user calibration dict
    using `type: 'DN'` as follows::

        calib_dict = {'B07': {'slope': 0.0037, 'offset': 18.5},
                      'B14': {'slope': -0.002, 'offset': 22.8},
                      'type': 'DN'}

    You can also explicitly select radiance correction with `'type': 'RAD'`
    but this is not necessary as it is the default option if you supply your
    own correction coefficients.

    """

    def __init__(self, filename, filename_info, filetype_info,
                 mask_space=True, calib_mode="update",
                 user_calibration=None, round_actual_position=True):
        """Initialize the reader."""
        super(AHIHSDFileHandler, self).__init__(filename, filename_info,
                                                filetype_info)

        # Register cleanup to be called when garbage collected
        weakref.finalize(self, self._cleanup)

        self.is_zipped = False
        self._unzipped = unzip_file(self.filename, prefix=str(filename_info["segment"]).zfill(2))
        # Assume file is not zipped
        if self._unzipped:
            # But if it is, set the filename to point to unzipped temp file
            self.is_zipped = True
            self.filename = self._unzipped

        self.channels = dict([(i, None) for i in AHI_CHANNEL_NAMES])
        self.units = dict([(i, "counts") for i in AHI_CHANNEL_NAMES])

        self._data = dict([(i, None) for i in AHI_CHANNEL_NAMES])
        self._header = dict([(i, None) for i in AHI_CHANNEL_NAMES])
        self.lons = None
        self.lats = None
        self.segment_number = filename_info["segment"]
        self.total_segments = filename_info["total_segments"]

        with open(self.filename) as fd:
            self.basic_info = np.fromfile(fd,
                                          dtype=_BASIC_INFO_TYPE,
                                          count=1)
            self.data_info = np.fromfile(fd,
                                         dtype=_DATA_INFO_TYPE,
                                         count=1)
            self.proj_info = np.fromfile(fd,
                                         dtype=_PROJ_INFO_TYPE,
                                         count=1)[0]
            self.nav_info = np.fromfile(fd,
                                        dtype=_NAV_INFO_TYPE,
                                        count=1)[0]
        self.platform_name = np2str(self.basic_info["satellite"])
        self.observation_area = np2str(self.basic_info["observation_area"])
        self.sensor = "ahi"
        self.mask_space = mask_space
        self.band_name = filetype_info["file_type"][4:].upper()
        calib_mode_choices = ("NOMINAL", "UPDATE")
        if calib_mode.upper() not in calib_mode_choices:
            raise ValueError("Invalid calibration mode: {}. Choose one of {}".format(
                calib_mode, calib_mode_choices))

        self.calib_mode = calib_mode.upper()
        self.user_calibration = user_calibration
        self._round_actual_position = round_actual_position

    def _cleanup(self):
        """Delete unzipped file."""
        if self.is_zipped and os.path.exists(self.filename):
            os.remove(self.filename)

    @property
    def start_time(self):
        """Get the nominal start time."""
        return self.nominal_start_time

    @property
    def end_time(self):
        """Get the nominal end time."""
        return self.nominal_end_time

    @property
    def observation_start_time(self):
        """Get the observation start time."""
        return dt.datetime(1858, 11, 17) + dt.timedelta(days=float(self.basic_info["observation_start_time"].item()))

    @property
    def observation_end_time(self):
        """Get the observation end time."""
        return dt.datetime(1858, 11, 17) + dt.timedelta(days=float(self.basic_info["observation_end_time"].item()))

    @property
    def _timeline(self):
        return "{:04d}".format(self.basic_info["observation_timeline"][0])

    @property
    def nominal_start_time(self):
        """Time this band was nominally to be recorded."""
        calc = _NominalTimeCalculator(self._timeline, self.observation_area)
        return calc.get_nominal_start_time(self.observation_start_time)

    @property
    def nominal_end_time(self):
        """Get the nominal end time."""
        calc = _NominalTimeCalculator(self._timeline, self.observation_area)
        return calc.get_nominal_end_time(self.nominal_start_time)

    def get_dataset(self, key, info):
        """Get the dataset."""
        return self.read_band(key, info)

    @cached_property
    def area(self):
        """Get AreaDefinition representing this file's data."""
        return self._get_area_def()

    def get_area_def(self, dsid):
        """Get the area definition."""
        del dsid
        return self.area

    def _get_area_def(self):
        pdict = {}
        pdict["cfac"] = np.uint32(self.proj_info["CFAC"])
        pdict["lfac"] = np.uint32(self.proj_info["LFAC"])
        pdict["coff"] = np.float32(self.proj_info["COFF"])
        pdict["loff"] = -np.float32(self.proj_info["LOFF"]) + 1
        pdict["a"] = float(self.proj_info["earth_equatorial_radius"] * 1000)
        pdict["h"] = float(self.proj_info["distance_from_earth_center"] * 1000 - pdict["a"])
        pdict["b"] = float(self.proj_info["earth_polar_radius"] * 1000)
        pdict["ssp_lon"] = float(self.proj_info["sub_lon"])
        pdict["nlines"] = int(self.data_info["number_of_lines"].item())
        pdict["ncols"] = int(self.data_info["number_of_columns"].item())
        pdict["scandir"] = "N2S"

        pdict["loff"] = pdict["loff"] + (self.segment_number * pdict["nlines"])

        aex = get_area_extent(pdict)

        pdict["a_name"] = self.observation_area
        pdict["a_desc"] = "AHI {} area".format(self.observation_area)
        pdict["p_id"] = f'geosh{self.basic_info["satellite"][0].decode()[-1]}'

        return get_area_definition(pdict, aex)

    def _check_fpos(self, fp_, fpos, offset, block):
        """Check file position matches blocksize."""
        if fp_.tell() + offset != fpos:
            warnings.warn(
                f"Actual {block} header size does not match expected",
                stacklevel=3
            )
        return

    def _read_header(self, fp_):
        """Read header."""
        header = {}

        fpos = 0
        header["block1"] = np.fromfile(
            fp_, dtype=_BASIC_INFO_TYPE, count=1)
        fpos = fpos + int(header["block1"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 0, "block1")
        fp_.seek(fpos, 0)
        header["block2"] = np.fromfile(fp_, dtype=_DATA_INFO_TYPE, count=1)
        fpos = fpos + int(header["block2"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 0, "block2")
        fp_.seek(fpos, 0)
        header["block3"] = np.fromfile(fp_, dtype=_PROJ_INFO_TYPE, count=1)
        fpos = fpos + int(header["block3"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 0, "block3")
        fp_.seek(fpos, 0)
        header["block4"] = np.fromfile(fp_, dtype=_NAV_INFO_TYPE, count=1)
        fpos = fpos + int(header["block4"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 0, "block4")
        fp_.seek(fpos, 0)
        header["block5"] = np.fromfile(fp_, dtype=_CAL_INFO_TYPE, count=1)
        logger.debug("Band number = " +
                     str(header["block5"]["band_number"][0]))
        logger.debug("Time_interval: %s - %s",
                     str(self.start_time), str(self.end_time))
        band_number = header["block5"]["band_number"][0]
        if band_number < 7:
            cal = np.fromfile(fp_, dtype=_VISCAL_INFO_TYPE, count=1)
        else:
            cal = np.fromfile(fp_, dtype=_IRCAL_INFO_TYPE, count=1)
        fpos = fpos + int(header["block5"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 0, "block5")
        fp_.seek(fpos, 0)

        header["calibration"] = cal

        header["block6"] = np.fromfile(
            fp_, dtype=_INTER_CALIBRATION_INFO_TYPE, count=1)
        fpos = fpos + int(header["block6"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 0, "block6")
        fp_.seek(fpos, 0)
        header["block7"] = np.fromfile(
            fp_, dtype=_SEGMENT_INFO_TYPE, count=1)
        fpos = fpos + int(header["block7"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 0, "block7")
        fp_.seek(fpos, 0)
        header["block8"] = np.fromfile(
            fp_, dtype=_NAVIGATION_CORRECTION_INFO_TYPE, count=1)
        # 8 The navigation corrections:
        ncorrs = header["block8"]["numof_correction_info_data"][0]
        corrections = []
        for _i in range(ncorrs):
            corrections.append(np.fromfile(fp_, dtype=_NAVIGATION_CORRECTION_SUBINFO_TYPE, count=1))
        fpos = fpos + int(header["block8"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 40, "block8")
        fp_.seek(fpos, 0)
        header["navigation_corrections"] = corrections
        header["block9"] = np.fromfile(fp_,
                                       dtype=_OBSERVATION_TIME_INFO_TYPE,
                                       count=1)
        numobstimes = header["block9"]["number_of_observation_times"][0]

        lines_and_times = []
        for _i in range(numobstimes):
            lines_and_times.append(np.fromfile(fp_,
                                               dtype=_OBSERVATION_LINE_TIME_INFO_TYPE,
                                               count=1))
        header["observation_time_information"] = lines_and_times
        fpos = fpos + int(header["block9"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 40, "block9")
        fp_.seek(fpos, 0)

        header["block10"] = np.fromfile(fp_,
                                        dtype=_ERROR_INFO_TYPE,
                                        count=1)
        num_err_info_data = header["block10"][
            "number_of_error_info_data"][0]
        err_info_data = []
        for _i in range(num_err_info_data):
            err_info_data.append(np.fromfile(fp_, dtype=_ERROR_LINE_INFO_TYPE, count=1))
        header["error_information_data"] = err_info_data
        fpos = fpos + int(header["block10"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 40, "block10")
        fp_.seek(fpos, 0)

        header["block11"] = np.fromfile(fp_, dtype=_SPARE_TYPE, count=1)
        fpos = fpos + int(header["block11"]["blocklength"].item())
        self._check_fpos(fp_, fpos, 0, "block11")
        fp_.seek(fpos, 0)

        return header

    def _read_data(self, fp_, header, resolution):
        """Read data block."""
        nlines = int(header["block2"]["number_of_lines"].item())
        ncols = int(header["block2"]["number_of_columns"].item())
        chunks = normalize_low_res_chunks(
            ("auto", "auto"),
            (nlines, ncols),
            # 1100 minimum chunk size for 500m, 550 for 1km, 225 for 2km
            (1100, 1100),
            (int(resolution / 500), int(resolution / 500)),
            np.float32,
        )
        return da.from_array(np.memmap(self.filename, offset=fp_.tell(),
                                       dtype="<u2", shape=(nlines, ncols), mode="r"),
                             chunks=chunks)

    def _mask_invalid(self, data, header):
        """Mask invalid data."""
        invalid = da.logical_or(data == header["block5"]["count_value_outside_scan_pixels"][0],
                                data == header["block5"]["count_value_error_pixels"][0])
        return da.where(invalid, np.float32(np.nan), data)

    def _mask_space(self, data):
        """Mask space pixels."""
        return data.where(get_geostationary_mask(self.area, chunks=data.chunks))

    def read_band(self, key, ds_info):
        """Read the data."""
        with open(self.filename, "rb") as fp_:
            self._header = self._read_header(fp_)
            res = self._read_data(fp_, self._header, key["resolution"])
        res = self._mask_invalid(data=res, header=self._header)
        res = self.calibrate(res, key["calibration"])

        new_info = self._get_metadata(key, ds_info)
        res = xr.DataArray(res, attrs=new_info, dims=["y", "x"])
        if self.mask_space:
            res = self._mask_space(res)
        return res

    def _get_metadata(self, key, ds_info):
        # Get actual satellite position. For altitude use the ellipsoid radius at the SSP.
        actual_lon = float(self.nav_info["SSP_longitude"])
        actual_lat = float(self.nav_info["SSP_latitude"])
        re = get_earth_radius(lon=actual_lon, lat=actual_lat,
                              a=float(self.proj_info["earth_equatorial_radius"] * 1000),
                              b=float(self.proj_info["earth_polar_radius"] * 1000))
        actual_alt = float(self.nav_info["distance_earth_center_to_satellite"]) * 1000 - re

        if self._round_actual_position:
            actual_lon = round(actual_lon, 3)
            actual_lat = round(actual_lat, 2)
            actual_alt = round(actual_alt / 150) * 150  # to the nearest 150m

        # Update metadata
        new_info = dict(
            units=ds_info["units"],
            standard_name=ds_info["standard_name"],
            wavelength=ds_info["wavelength"],
            resolution=ds_info["resolution"],
            id=key,
            name=key["name"],
            platform_name=self.platform_name,
            sensor=self.sensor,
            time_parameters=dict(
                nominal_start_time=self.nominal_start_time,
                nominal_end_time=self.nominal_end_time,
                observation_start_time=self.observation_start_time,
                observation_end_time=self.observation_end_time,
            ),
            orbital_parameters={
                "projection_longitude": float(self.proj_info["sub_lon"]),
                "projection_latitude": 0.,
                "projection_altitude": float(self.proj_info["distance_from_earth_center"] -
                                             self.proj_info["earth_equatorial_radius"]) * 1000,
                "satellite_actual_longitude": actual_lon,
                "satellite_actual_latitude": actual_lat,
                "satellite_actual_altitude": actual_alt,
                "nadir_longitude": float(self.nav_info["nadir_longitude"]),
                "nadir_latitude": float(self.nav_info["nadir_latitude"]),
            },
        )
        return new_info

    def calibrate(self, data, calibration):
        """Calibrate the data."""
        if calibration == "counts":
            return data

        if calibration in ["radiance", "reflectance", "brightness_temperature"]:
            data = self.convert_to_radiance(data)
        if calibration == "reflectance":
            data = self._vis_calibrate(data)
        elif calibration == "brightness_temperature":
            data = self._ir_calibrate(data)
        return data

    def convert_to_radiance(self, data):
        """Calibrate to radiance."""
        bnum = self._header["block5"]["band_number"][0]
        # Check calibration mode and select corresponding coefficients
        if self.calib_mode == "UPDATE" and bnum < 7:
            dn_gain = self._header["calibration"]["cali_gain_count2rad_conversion"][0]
            dn_offset = self._header["calibration"]["cali_offset_count2rad_conversion"][0]
            if dn_gain == 0 and dn_offset == 0:
                logger.info(
                    "No valid updated coefficients, fall back to default values.")
                dn_gain = self._header["block5"]["gain_count2rad_conversion"][0]
                dn_offset = self._header["block5"]["offset_count2rad_conversion"][0]
        else:
            dn_gain = self._header["block5"]["gain_count2rad_conversion"][0]
            dn_offset = self._header["block5"]["offset_count2rad_conversion"][0]

        # Assume no user correction
        correction_type = self._get_user_calibration_correction_type()
        if correction_type == "DN":
            # Replace file calibration with user calibration
            dn_gain, dn_offset = get_user_calibration_factors(self.band_name,
                                                              self.user_calibration)

        data = (data * np.float32(dn_gain) + np.float32(dn_offset))
        # If using radiance correction factors from GSICS or similar, apply here
        if correction_type == "RAD":
            user_slope, user_offset = get_user_calibration_factors(self.band_name,
                                                                   self.user_calibration)
            data = apply_rad_correction(data, np.float32(user_slope), np.float32(user_offset))
        return data

    def _get_user_calibration_correction_type(self):
        correction_type = None
        if isinstance(self.user_calibration, dict):
            # Check if we have DN correction coeffs
            correction_type = self.user_calibration.get("type", "RAD")
        return correction_type

    def _vis_calibrate(self, data):
        """Visible channel calibration only."""
        coeff = self._header["calibration"]["coeff_rad2albedo_conversion"]
        return (data * np.float32(coeff) * 100).clip(0)

    def _ir_calibrate(self, data):
        """IR calibration."""
        # No radiance -> no temperature
        data = da.where(data == 0, np.float32(np.nan), data)

        cwl = self._header["block5"]["central_wave_length"][0] * 1e-6
        c__ = self._header["calibration"]["speed_of_light"][0]
        h__ = self._header["calibration"]["planck_constant"][0]
        k__ = self._header["calibration"]["boltzmann_constant"][0]
        a__ = (h__ * c__) / (k__ * cwl)

        b__ = ((2 * h__ * c__ ** 2) / (data * 1.0e6 * cwl ** 5)) + 1

        Te_ = a__ / da.log(b__)

        c0_ = self._header["calibration"]["c0_rad2tb_conversion"][0]
        c1_ = self._header["calibration"]["c1_rad2tb_conversion"][0]
        c2_ = self._header["calibration"]["c2_rad2tb_conversion"][0]

        return (c0_ + c1_ * Te_ + c2_ * Te_ ** 2).clip(0)


class _NominalTimeCalculator:
    """Get time when a scan was nominally to be recorded."""

    def __init__(self, timeline, area):
        """Initialize the nominal timestamp calculator.

        Args:
            timeline (str): Observation timeline (four characters HHMM)
            area (str): Observation area (four characters, e.g. FLDK)
        """
        self.timeline = self._parse_timeline(timeline)
        self.area = area

    def _parse_timeline(self, timeline):
        try:
            return dt.datetime.strptime(timeline, "%H%M").time()
        except ValueError:
            return None

    def get_nominal_start_time(self, observation_start_time):
        """Get nominal start time of the scan."""
        return self._modify_observation_time_for_nominal(observation_start_time)

    def get_nominal_end_time(self, nominal_start_time):
        """Get nominal end time of the scan."""
        freq = self._observation_frequency
        return nominal_start_time + dt.timedelta(minutes=freq // 60,
                                                 seconds=freq % 60)

    def _modify_observation_time_for_nominal(self, observation_time):
        """Round observation time to a nominal time based on known observation frequency.

        AHI observations are split into different sectors including Full Disk
        (FLDK), Japan (JP) sectors, and smaller regional (R) sectors. Each
        sector is observed at different frequencies (ex. every 10 minutes,
        every 2.5 minutes, and every 30 seconds). This method will take the
        actual observation time and round it to the nearest interval for this
        sector. So if the observation time is 13:32:48 for the "JP02" sector
        which is the second Japan observation where every Japan observation is
        2.5 minutes apart, then the result should be 13:32:30.
        """
        if not self.timeline:
            warnings.warn(
                "Observation timeline is fill value, not rounding observation time.",
                stacklevel=3
            )
            return observation_time
        timeline = self._get_closest_timeline(observation_time)
        offset = self._get_offset_relative_to_timeline()
        return timeline + dt.timedelta(minutes=offset//60, seconds=offset % 60)

    def _get_closest_timeline(self, observation_time):
        """Find the closest timeline for the given observation time.

        Needs to check surrounding days because the observation might start
        a little bit before the planned time.

        Observation start time: 2022-12-31 23:59
        Timeline: 0000
        => Nominal start time: 2023-01-01 00:00
        """
        delta_days = [-1, 0, 1]
        surrounding_dates = [
            (observation_time + dt.timedelta(days=delta)).date()
            for delta in delta_days
        ]
        timelines = [
            dt.datetime.combine(date, self.timeline)
            for date in surrounding_dates
        ]
        diffs = [
            abs((timeline - observation_time))
            for timeline in timelines
        ]
        argmin = np.argmin(diffs)
        return timelines[argmin]

    def _get_offset_relative_to_timeline(self):
        if self.area == "FLDK":
            return 0
        sector_repeat = int(self.area[2:]) - 1
        return self._observation_frequency * sector_repeat

    @property
    def _observation_frequency(self):
        frequencies = {"FLDK": 600, "JP": 150, "R3": 150, "R4": 30, "R5": 30}
        area = self.area
        if area != "FLDK":
            # e.g. JP01, JP02 etc
            area = area[:2]
        return frequencies[area]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""EUMETSAT base reader tests package."""

import datetime as dt
import unittest

import numpy as np

from satpy.readers.core.eum import (
    get_service_mode,
    recarray2dict,
    time_cds,
    time_cds_expanded,
    time_cds_short,
    timecds2datetime,
)
from satpy.readers.core.seviri import mpef_product_header


class TestMakeTimeCdsDictionary(unittest.TestCase):
    """Test TestMakeTimeCdsDictionary."""

    def test_fun(self):
        """Test function for TestMakeTimeCdsDictionary."""
        # time_cds_short
        tcds = {"Days": np.array(1), "Milliseconds": np.array(2)}
        expected = dt.datetime(1958, 1, 2, 0, 0, 0, 2000)
        assert timecds2datetime(tcds) == expected

        # time_cds
        tcds = {"Days": np.array(1), "Milliseconds": np.array(2), "Microseconds": np.array(3)}
        expected = dt.datetime(1958, 1, 2, 0, 0, 0, 2003)
        assert timecds2datetime(tcds) == expected

        # time_cds_expanded
        tcds = {"Days": np.array(1), "Milliseconds": np.array(2), "Microseconds": np.array(3),
                "Nanoseconds": np.array(4)}
        expected = dt.datetime(1958, 1, 2, 0, 0, 0, 2003)
        assert timecds2datetime(tcds) == expected


class TestMakeTimeCdsRecarray(unittest.TestCase):
    """Test TestMakeTimeCdsRecarray."""

    def test_fun(self):
        """Test function for TestMakeTimeCdsRecarray."""
        # time_cds_short
        tcds = np.array([(1, 2)], dtype=np.dtype(time_cds_short))
        expected = dt.datetime(1958, 1, 2, 0, 0, 0, 2000)
        assert timecds2datetime(tcds) == expected

        # time_cds
        tcds = np.array([(1, 2, 3)], dtype=np.dtype(time_cds))
        expected = dt.datetime(1958, 1, 2, 0, 0, 0, 2003)
        assert timecds2datetime(tcds) == expected

        # time_cds_expanded
        tcds = np.array([(1, 2, 3, 4)], dtype=np.dtype(time_cds_expanded))
        expected = dt.datetime(1958, 1, 2, 0, 0, 0, 2003)
        assert timecds2datetime(tcds) == expected


class TestRecarray2Dict(unittest.TestCase):
    """Test TestRecarray2Dict."""

    def test_timestamps(self):
        """Test function for TestRecarray2Dict."""
        # datatype definition

        pat_dt = np.dtype([
            ("TrueRepeatCycleStart", time_cds_expanded),
            ("PlanForwardScanEnd", time_cds_expanded),
            ("PlannedRepeatCycleEnd", time_cds_expanded)
        ])

        # planned acquisition time, add extra dimensions
        # these should be removed by recarray2dict
        pat = np.array([[[(
            (21916, 41409544, 305, 262),
            (21916, 42160340, 659, 856),
            (21916, 42309417, 918, 443))]]], dtype=pat_dt)

        expected = {
            "TrueRepeatCycleStart": dt.datetime(2018, 1, 2, 11, 30, 9, 544305),
            "PlanForwardScanEnd": dt.datetime(2018, 1, 2, 11, 42, 40, 340660),
            "PlannedRepeatCycleEnd": dt.datetime(2018, 1, 2, 11, 45, 9, 417918)
        }

        assert recarray2dict(pat) == expected

    def test_mpef_product_header(self):
        """Test function for TestRecarray2Dict and mpef product header."""
        names = ["ImageLocation", "GsicsCalMode", "GsicsCalValidity",
                 "Padding", "OffsetToData", "Padding2"]
        mpef_header = np.dtype([(name, mpef_product_header.fields[name][0])
                                for name in names])
        mph_struct = np.array([("OPE", True, False, "XX", 1000, "12345678")], dtype=mpef_header)
        test_mph = {"ImageLocation": "OPE",
                    "GsicsCalMode": True,
                    "GsicsCalValidity": False,
                    "Padding": "XX",
                    "OffsetToData": 1000,
                    "Padding2": "12345678"
                    }
        assert recarray2dict(mph_struct) == test_mph


class TestGetServiceMode(unittest.TestCase):
    """Test the get_service_mode function."""

    def test_get_seviri_service_mode_fes(self):
        """Test fetching of SEVIRI service mode information for FES."""
        ssp_lon = 0.0
        name = "fes"
        desc = "Full Earth Scanning service"
        res = get_service_mode("seviri", ssp_lon)
        assert res["service_name"] == name
        assert res["service_desc"] == desc

    def test_get_seviri_service_mode_rss(self):
        """Test fetching of SEVIRI service mode information for RSS."""
        ssp_lon = 9.5
        name = "rss"
        desc = "Rapid Scanning Service"
        res = get_service_mode("seviri", ssp_lon)
        assert res["service_name"] == name
        assert res["service_desc"] == desc

    def test_get_seviri_service_mode_iodc_E0415(self):
        """Test fetching of SEVIRI service mode information for IODC at 41.5 degrees East."""
        ssp_lon = 41.5
        name = "iodc"
        desc = "Indian Ocean Data Coverage service"
        res = get_service_mode("seviri", ssp_lon)
        assert res["service_name"] == name
        assert res["service_desc"] == desc

    def test_get_seviri_service_mode_iodc_E0455(self):
        """Test fetching of SEVIRI service mode information for IODC at 45.5 degrees East."""
        ssp_lon = 45.5
        name = "iodc"
        desc = "Indian Ocean Data Coverage service"
        res = get_service_mode("seviri", ssp_lon)
        assert res["service_name"] == name
        assert res["service_desc"] == desc

    def test_get_fci_service_mode_fdss(self):
        """Test fetching of FCI service mode information for FDSS."""
        ssp_lon = 0.0
        name = "fdss"
        desc = "Full Disk Scanning Service"
        res = get_service_mode("fci", ssp_lon)
        assert res["service_name"] == name
        assert res["service_desc"] == desc

    def test_get_fci_service_mode_rss(self):
        """Test fetching of FCI service mode information for RSS."""
        ssp_lon = 9.5
        name = "rss"
        desc = "Rapid Scanning Service"
        res = get_service_mode("fci", ssp_lon)
        assert res["service_name"] == name
        assert res["service_desc"] == desc

    def test_get_unknown_lon_service_mode(self):
        """Test fetching of service mode information for unknown input longitude."""
        ssp_lon = 13
        name = "unknown"
        desc = "unknown"
        res = get_service_mode("fci", ssp_lon)
        assert res["service_name"] == name
        assert res["service_desc"] == desc

    def test_get_unknown_instrument_service_mode(self):
        """Test fetching of service mode information for unknown input instrument."""
        ssp_lon = 0
        name = "unknown"
        desc = "unknown"
        res = get_service_mode("test", ssp_lon)
        assert res["service_name"] == name
        assert res["service_desc"] == desc
